﻿using Microsoft.EntityFrameworkCore;
using Microsoft.EntityFrameworkCore.Metadata.Builders;
using System;
using System.ComponentModel;
using System.Reflection;
using Volo.Abp.Auditing;
using Volo.Abp.Data;
using Volo.Abp.Domain.Entities;
using Volo.Abp.MultiTenancy;

namespace Volo.Abp.EntityFrameworkCore.Modeling;

public static class EntityTypeBuilderExtensions
{
    /// <summary>
    /// 根据<see cref="DisplayNameAttribute.DisplayName"/>设置字段注释。
    /// </summary>
    /// <typeparam name="T">实体</typeparam>
    /// <param name="builder"></param>
    public static void ConfigureCommentByDisplayName<T>(this EntityTypeBuilder<T> builder) where T : class
    {
        Check.NotNull(builder, nameof(builder));
        var type = typeof(T);
        var properties = type.GetProperties();
        foreach (var propertyInfo in properties)
        {
            var text = propertyInfo.GetCustomAttribute<DisplayNameAttribute>()?.DisplayName;
            if (!string.IsNullOrEmpty(text))
            {
                builder.Property(propertyInfo.Name).HasComment(text);
            }
        }
        builder.HasComment(type.GetCustomAttribute<DisplayNameAttribute>()?.DisplayName);
    }

    public static void ConfigureByConventionWithChineseComment(this EntityTypeBuilder b)
    {
        AbpEntityTypeBuilderExtensions.ConfigureByConvention(b);

        if (b.Metadata.ClrType.IsAssignableTo<IHasConcurrencyStamp>())
        {
            b.Property(nameof(IHasConcurrencyStamp.ConcurrencyStamp)).HasComment("并发标识");
        }

        if (b.Metadata.ClrType.IsAssignableTo<IHasExtraProperties>())
        {
            b.Property<ExtraPropertyDictionary>(nameof(IHasExtraProperties.ExtraProperties)).HasComment("扩展属性");
        }

        if (b.Metadata.ClrType.IsAssignableTo<IMayHaveCreator>())
        {
            b.Property(nameof(IMayHaveCreator.CreatorId)).HasComment("创建人Id");
        }

        if (b.Metadata.ClrType.IsAssignableTo<ISoftDelete>())
        {
            b.Property(nameof(ISoftDelete.IsDeleted)).HasComment("是否删除");
        }

        if (b.Metadata.ClrType.IsAssignableTo<IHasDeletionTime>())
        {
            b.Property(nameof(IHasDeletionTime.DeletionTime)).HasComment("删除时间");
        }

        if (b.Metadata.ClrType.IsAssignableTo<IDeletionAuditedObject>())
        {
            b.Property(nameof(IDeletionAuditedObject.DeleterId)).HasComment("删除人Id");
        }

        if (b.Metadata.ClrType.IsAssignableTo<IHasCreationTime>())
        {
            b.Property(nameof(IHasCreationTime.CreationTime)).HasComment("创建时间");
        }

        if (b.Metadata.ClrType.IsAssignableTo<IHasModificationTime>())
        {
            b.Property(nameof(IHasModificationTime.LastModificationTime)).HasComment("修改时间");
        }

        if (b.Metadata.ClrType.IsAssignableTo<IModificationAuditedObject>())
        {
            b.Property(nameof(IModificationAuditedObject.LastModifierId)).HasComment("最后修改人Id");
        }

        if (b.Metadata.ClrType.IsAssignableTo<IMultiTenant>())
        {
            b.Property(nameof(IMultiTenant.TenantId)).HasComment("租户Id");
        }
    }
}
